@@ -40,9 +40,9 @@
 
 #include <string.h>
 
-#include "galias.h"
 #include "glib.h"
 #include "gthreadinit.h"
+#include "galias.h"
 
 #if GLIB_SIZEOF_SYSTEM_THREAD == SIZEOF_VOID_P
 # define g_system_thread_equal_simple(thread1, thread2)			\
@@ -152,6 +152,7 @@ g_thread_init_glib (void)
   _g_mem_thread_init ();
   _g_messages_thread_init ();
   _g_atomic_thread_init ();
+  _g_utils_thread_init ();
 #ifdef G_OS_WIN32
   _g_win32_thread_init ();
 #endif
@@ -334,6 +335,9 @@ g_static_rec_mutex_lock_full   (GStaticRecMutex *mutex,
   if (!g_thread_supported ())
     return;
 
+  if (depth == 0)
+    return;
+
   G_THREAD_UF (thread_self, (&self));
 
   if (g_system_thread_equal (self, mutex->owner))
@@ -836,3 +840,6 @@ g_static_rw_lock_free (GStaticRWLock* lock)
     }
   g_static_mutex_free (&lock->mutex);
 }
+
+#define __G_THREAD_C__
+#include "galiasdef.c"
