@@ -23,12 +23,13 @@
 
 #include        <string.h>
 
-#include	"gobjectalias.h"
 #include	"genums.h"
 
 #include	"gvalue.h"
 #include	"gvaluecollector.h"
 
+#include	"gobjectalias.h"
+
 
 /* --- prototypes --- */
 static void	g_enum_class_init		(GEnumClass	*class,
@@ -378,10 +379,19 @@ g_flags_get_first_value (GFlagsClass *flags_class,
   if (flags_class->n_values)
     {
       GFlagsValue *flags_value;
-      
-      for (flags_value = flags_class->values; flags_value->value_name; flags_value++)
-	if ((flags_value->value & value) == flags_value->value)
-	  return flags_value;
+
+      if (value == 0)
+        {
+          for (flags_value = flags_class->values; flags_value->value_name; flags_value++)
+            if (flags_value->value == 0)
+              return flags_value;
+        }
+      else
+        {
+          for (flags_value = flags_class->values; flags_value->value_name; flags_value++)
+            if (flags_value->value != 0 && (flags_value->value & value) == flags_value->value)
+              return flags_value;
+        }      
     }
   
   return NULL;
@@ -420,3 +430,6 @@ g_value_get_flags (const GValue *value)
   
   return value->data[0].v_ulong;
 }
+
+#define __G_ENUMS_C__
+#include "gobjectaliasdef.c"
